@@ -1,9 +1,7 @@
 # Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
 """Pure Python implementation of DLT library"""
-# pylint: disable=too-many-lines
-from __future__ import absolute_import
-
 import ctypes
+import ipaddress as ip
 import logging
 import os
 import re
@@ -12,34 +10,48 @@ import struct
 import time
 import threading
 
-import six
-
 from dlt.core import (
-    dltlib, DLT_ID_SIZE, DLT_HTYP_WEID, DLT_HTYP_WSID, DLT_HTYP_WTMS, DLT_HTYP_UEH, DLT_RETURN_OK,
-    DLT_RETURN_ERROR, DLT_RETURN_TRUE, DLT_FILTER_MAX, DLT_MESSAGE_ERROR_OK, cDltExtendedHeader,
-    cDltClient, MessageMode, cDLTMessage, cDltStorageHeader, cDltStandardHeader,
-    DLT_TYPE_INFO_UINT, DLT_TYPE_INFO_SINT, DLT_TYPE_INFO_STRG, DLT_TYPE_INFO_SCOD,
-    DLT_TYPE_INFO_TYLE, DLT_TYPE_INFO_VARI, DLT_TYPE_INFO_RAWD,
-    DLT_SCOD_ASCII, DLT_SCOD_UTF8, DLT_TYLE_8BIT, DLT_TYLE_16BIT, DLT_TYLE_32BIT, DLT_TYLE_64BIT,
-    DLT_TYLE_128BIT, DLT_DAEMON_TCP_PORT, DLT_CLIENT_RCVBUFSIZE,
+    cDLTFilter,
+    dltlib,
+    DLT_CLIENT_MODE_UDP_MULTICAST,
+    DLT_ID_SIZE,
+    DLT_HTYP_WEID,
+    DLT_HTYP_WSID,
+    DLT_HTYP_WTMS,
+    DLT_HTYP_UEH,
+    DLT_RETURN_OK,
+    DLT_RETURN_ERROR,
+    DLT_RETURN_TRUE,
+    DLT_MESSAGE_ERROR_OK,
+    cDltExtendedHeader,
+    cDltClient,
+    MessageMode,
+    cDLTMessage,
+    cDltStorageHeader,
+    cDltStandardHeader,
+    DLT_TYPE_INFO_UINT,
+    DLT_TYPE_INFO_SINT,
+    DLT_TYPE_INFO_STRG,
+    DLT_TYPE_INFO_SCOD,
+    DLT_TYPE_INFO_TYLE,
+    DLT_TYPE_INFO_VARI,
+    DLT_TYPE_INFO_RAWD,
+    DLT_SCOD_ASCII,
+    DLT_SCOD_UTF8,
+    DLT_TYLE_8BIT,
+    DLT_TYLE_16BIT,
+    DLT_TYLE_32BIT,
+    DLT_TYLE_64BIT,
+    DLT_TYLE_128BIT,
+    DLT_DAEMON_TCP_PORT,
+    DLT_CLIENT_RCVBUFSIZE,
     DLT_RECEIVE_SOCKET,
 )
 from dlt.helpers import bytes_to_str
 
-try:
-    # Use xrange by default on Python 2
-    range = xrange  # pylint: disable=redefined-builtin,undefined-variable,invalid-name
-except Exception:  # pylint: disable=broad-except
-    pass
-
 MAX_LOG_IN_ROW = 3
-# Return value for DLTFilter.add() - exceeded maximum number of filters
-MAX_FILTER_REACHED = 1
-# Return value for DLTFilter.add() - specified filter already exists
-REPEATED_FILTER = 2
 logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
 
-
 DLT_EMPTY_FILE_ERROR = "DLT TRACE FILE IS EMPTY"
 cDLT_FILE_NOT_OPEN_ERROR = "Could not open DLT Trace file (libdlt)"  # pylint: disable=invalid-name
 
@@ -54,7 +66,7 @@ class cached_property(object):  # pylint: disable=invalid-name
     """  # noqa
 
     def __init__(self, func):
-        self.__doc__ = getattr(func, '__doc__')
+        self.__doc__ = getattr(func, "__doc__")
         self.func = func
 
     def __get__(self, obj, cls):
@@ -64,22 +76,11 @@ class cached_property(object):  # pylint: disable=invalid-name
         return value
 
 
-class DLTFilter(ctypes.Structure):
+class DLTFilter(cDLTFilter):
     """Structure to store filter parameters. ID are maximal four characters. Unused values are filled with zeros.
     If every value as filter is valid, the id should be empty by having only zero values.
-
-    typedef struct
-    {
-        char apid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< application id */
-        char ctid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< context id */
-        int  counter;                           /**< number of filters */
-    } DltFilter;
     """
 
-    _fields_ = [("apid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
-                ("ctid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
-                ("counter", ctypes.c_int)]
-
     verbose = 0
 
     def __init__(self, **kwords):
@@ -92,27 +93,12 @@ class DLTFilter(ctypes.Structure):
         if dltlib.dlt_filter_free(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
             raise RuntimeError("Could not cleanup DLTFilter")
 
-    def add(self, apid, ctid):
-        """Add new filter pair"""
-        if six.PY3:
-            if isinstance(apid, str):
-                apid = bytes(apid, "ascii")
-            if isinstance(ctid, str):
-                ctid = bytes(ctid, "ascii")
-        if dltlib.dlt_filter_add(ctypes.byref(self), apid or b"", ctid or b"", self.verbose) == DLT_RETURN_ERROR:
-            if self.counter >= DLT_FILTER_MAX:
-                logger.error("Maximum number (%d) of allowed filters reached, ignoring filter!\n", DLT_FILTER_MAX)
-                return MAX_FILTER_REACHED
-            logger.debug("Filter ('%s', '%s') already exists", apid, ctid)
-            return REPEATED_FILTER
-        return 0
-
     def __repr__(self):
         """return the 'official' string representation of an object"""
         apids = [ctypes.string_at(entry[:DLT_ID_SIZE]) for entry in self.apid]
         ctids = [ctypes.string_at(entry[:DLT_ID_SIZE]) for entry in self.ctid]
 
-        return str(list(zip(apids[:self.counter], ctids[:self.counter])))
+        return str(list(zip(apids[: self.counter], ctids[: self.counter])))
 
     def __nonzero__(self):
         """Truth value testing"""
@@ -158,10 +144,9 @@ class Payload(object):
             value = None
             if type_info & DLT_TYPE_INFO_STRG:
                 if (get_scod(type_info) == DLT_SCOD_ASCII) or (get_scod(type_info) == DLT_SCOD_UTF8):
-
                     length = struct.unpack_from("H", self._buf, offset)[0]
                     offset += struct.calcsize("H")
-                    value = self._buf[offset:offset + length - 1]  # strip the string terminating char \x00
+                    value = self._buf[offset : offset + length - 1]  # strip the string terminating char \x00
                     offset += length
 
             elif type_info & DLT_TYPE_INFO_UINT:
@@ -214,7 +199,7 @@ class Payload(object):
                 length = struct.unpack_from("H", self._buf, offset)[0]
                 offset += struct.calcsize("H")
 
-                value = self._buf[offset:offset + length]
+                value = self._buf[offset : offset + length]
                 offset += length
 
             else:
@@ -238,6 +223,8 @@ class DLTMessage(cDLTMessage, MessageMode):
     # object is not initialized if the message is loaded from a file
     initialized_as_object = False
 
+    re_pattern_type = type(re.compile(r"type"))
+
     def __init__(self, *args, **kwords):
         self.initialized_as_object = True
         self.verbose = kwords.pop("verbose", 0)
@@ -261,13 +248,15 @@ class DLTMessage(cDLTMessage, MessageMode):
         ctypes.memmove(databuffer, self.databuffer, self.datasize)
 
         init_args = (self.found_serialheader, self.resync_offset, self.headersize, self.datasize)
-        state_dict = {'headerbuffer': bytearray(self.headerbuffer),
-                      'databuffer': bytearray(databuffer),
-                      'databuffersize': self.databuffersize,
-                      'storageheader': self.storageheader,
-                      'standardheader': self.standardheader,
-                      'headerextra': self.headerextra,
-                      'extendedheader': self.extendedheader, }
+        state_dict = {
+            "headerbuffer": bytearray(self.headerbuffer),
+            "databuffer": bytearray(databuffer),
+            "databuffersize": self.databuffersize,
+            "storageheader": self.storageheader,
+            "standardheader": self.standardheader,
+            "headerextra": self.headerextra,
+            "extendedheader": self.extendedheader,
+        }
         return (DLTMessage, init_args, state_dict)
 
     # pylint: disable=attribute-defined-outside-init
@@ -277,19 +266,19 @@ class DLTMessage(cDLTMessage, MessageMode):
         This method is called by the pickle module to populate a
         deserialized object's state after it has been created.
         """
-        self.databuffersize = state['databuffersize']
-        self.p_storageheader.contents = state['storageheader']
-        self.p_standardheader.contents = state['standardheader']
-        self.headerextra = state['headerextra']
-        self.p_extendedheader.contents = state['extendedheader']
+        self.databuffersize = state["databuffersize"]
+        self.p_storageheader.contents = state["storageheader"]
+        self.p_standardheader.contents = state["standardheader"]
+        self.headerextra = state["headerextra"]
+        self.p_extendedheader.contents = state["extendedheader"]
         # - populate databuffer
         databuffer = ctypes.ARRAY(ctypes.c_uint8, self.datasize)()
-        for index, byte in enumerate(state['databuffer']):
+        for index, byte in enumerate(state["databuffer"]):
             databuffer[index] = byte
         self.databuffer = databuffer
 
         # - populate headerbuffer
-        for index, byte in enumerate(state['headerbuffer']):
+        for index, byte in enumerate(state["headerbuffer"]):
             self.headerbuffer[index] = byte
 
         # - This is required because we are not calling
@@ -306,11 +295,13 @@ class DLTMessage(cDLTMessage, MessageMode):
 
         buf = ctypes.create_string_buffer(remainder)
 
-        dltlib.dlt_message_read(ctypes.byref(msg),
-                                ctypes.cast(buf, ctypes.POINTER(ctypes.c_uint8)),
-                                ctypes.c_uint(len(remainder)),
-                                0,  # resync
-                                0)  # verbose
+        dltlib.dlt_message_read(
+            ctypes.byref(msg),
+            ctypes.cast(buf, ctypes.POINTER(ctypes.c_uint8)),
+            ctypes.c_uint(len(remainder)),
+            0,  # resync
+            0,
+        )  # verbose
         msg.p_storageheader.contents = storageheader
         msg.initialized_as_object = False
 
@@ -327,21 +318,21 @@ class DLTMessage(cDLTMessage, MessageMode):
     @staticmethod
     def extract_storageheader(data):
         """Split binary message data into storage header and remainder"""
-        header = data[0:ctypes.sizeof(cDltStorageHeader)]
+        header = data[0 : ctypes.sizeof(cDltStorageHeader)]
         # pylint: disable=no-member
-        return (cDltStorageHeader.from_buffer_copy(header), data[ctypes.sizeof(cDltStorageHeader):])
+        return (cDltStorageHeader.from_buffer_copy(header), data[ctypes.sizeof(cDltStorageHeader) :])
 
     @staticmethod
     def extract_sort_data(data):
         """Extract timestamp, message length, apid, ctid from a bytestring in DLT storage format (speed optimized)"""
-        htyp_data = ord(chr(data[16])) if six.PY3 else ord(data[16])
+        htyp_data = ord(chr(data[16]))
         len_data = data[19:17:-1]
         len_value = ctypes.cast(len_data, ctypes.POINTER(ctypes.c_ushort)).contents.value + 16
         apid = b""
         ctid = b""
         tmsp_value = 0.0
 
-        bytes_offset = 0    # We know where data will be in the message, but ...
+        bytes_offset = 0  # We know where data will be in the message, but ...
         if not htyp_data & DLT_HTYP_WEID:  # if there is no ECU ID and/or Session ID, then it will be earlier
             bytes_offset -= 4
         if not htyp_data & DLT_HTYP_WSID:
@@ -349,13 +340,13 @@ class DLTMessage(cDLTMessage, MessageMode):
 
         if htyp_data & DLT_HTYP_WTMS:
             tmsp_base = 31 + bytes_offset  # Typical timestamp end offset
-            tmsp_data = data[tmsp_base:tmsp_base - 4:-1]
+            tmsp_data = data[tmsp_base : tmsp_base - 4 : -1]
             tmsp_value = ctypes.cast(tmsp_data, ctypes.POINTER(ctypes.c_uint32)).contents.value / 10000.0
 
         if htyp_data & DLT_HTYP_UEH:
             apid_base = 38 + bytes_offset  # Typical APID end offset
-            apid = data[apid_base - 4:apid_base].rstrip(b"\x00")
-            ctid = data[apid_base:apid_base + 4].rstrip(b"\x00")
+            apid = data[apid_base - 4 : apid_base].rstrip(b"\x00")
+            ctid = data[apid_base : apid_base + 4].rstrip(b"\x00")
 
         apid = bytes_to_str(apid)
         ctid = bytes_to_str(ctid)
@@ -390,8 +381,8 @@ class DLTMessage(cDLTMessage, MessageMode):
 
     def __eq__(self, other):
         """Equal test - not comparing storage header (contains timestamps)"""
-        header1 = ctypes.string_at(self.headerbuffer, self.headersize)[ctypes.sizeof(cDltStorageHeader):]
-        header2 = ctypes.string_at(other.headerbuffer, other.headersize)[ctypes.sizeof(cDltStorageHeader):]
+        header1 = ctypes.string_at(self.headerbuffer, self.headersize)[ctypes.sizeof(cDltStorageHeader) :]
+        header2 = ctypes.string_at(other.headerbuffer, other.headersize)[ctypes.sizeof(cDltStorageHeader) :]
 
         data1 = ctypes.string_at(self.databuffer, self.datasize)
         data2 = ctypes.string_at(other.databuffer, other.datasize)
@@ -420,23 +411,30 @@ class DLTMessage(cDLTMessage, MessageMode):
             return self.apid == other.apid and self.ctid == other.ctid and self.__eq__(other)
 
         # pylint: disable=protected-access
-        if hasattr(other, "_fields_") and [x[0] for x in other._fields_] == ["apid", "ctid", "counter"]:
+        if hasattr(other, "_fields_") and [x[0] for x in other._fields_] == [
+            "apid",
+            "ctid",
+            "log_level",
+            "payload_max",
+            "payload_min",
+            "counter",
+        ]:
             # other id DLTFilter
             return dltlib.dlt_message_filter_check(ctypes.byref(self), ctypes.byref(other), 0)
 
         if not isinstance(other, dict):
-            raise TypeError("other must be instance of mgu_dlt.dlt.DLTMessage, mgu_dlt.dlt.DLTFilter or a dictionary"
-                            " found: {}".format(type(other)))
-
-        re_pattern_type = type(re.compile(r"type"))
+            raise TypeError(
+                "other must be instance of dlt.dlt.DLTMessage, dlt.dlt.DLTFilter or a dictionary"
+                " found: {}".format(type(other))
+            )
 
         other = other.copy()
         apid = other.get("apid", None)
-        if apid and not isinstance(apid, re_pattern_type) and self.apid != apid:
+        if apid and not isinstance(apid, self.re_pattern_type) and self.apid != apid:
             return False
 
         ctid = other.get("ctid", None)
-        if ctid and not isinstance(ctid, re_pattern_type) and self.ctid != ctid:
+        if ctid and not isinstance(ctid, self.re_pattern_type) and self.ctid != ctid:
             return False
 
         for key, val in other.items():
@@ -446,7 +444,7 @@ class DLTMessage(cDLTMessage, MessageMode):
             msg_val = getattr(self, key, b"")
             if not msg_val:
                 return False
-            if isinstance(val, re_pattern_type):
+            if isinstance(val, self.re_pattern_type):
                 if not val.search(msg_val):
                     return False
             elif msg_val != val:
@@ -469,7 +467,7 @@ class DLTMessage(cDLTMessage, MessageMode):
     # convenient access to import DLT message attributes
     # no need to remember in which header are those attrs defined
     @cached_property
-    def ecuid(self):   # pylint: disable=invalid-overridden-method
+    def ecuid(self):  # pylint: disable=invalid-overridden-method
         """Get the ECU ID
 
         :returns: ECU ID
@@ -558,7 +556,7 @@ class DLTMessage(cDLTMessage, MessageMode):
         :returns: storage header timestamp
         :rtype: float
         """
-        return float("{}.{}".format(self.storageheader.seconds, self.storageheader.microseconds))
+        return self.storageheader.seconds + self.storageheader.microseconds * 0.000001
 
 
 class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
@@ -574,8 +572,8 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
         int32_t counter;       /**< number of messages in DLT file with filter */
         int32_t counter_total; /**< number of messages in DLT file without filter */
         int32_t position;      /**< current index to message parsed in DLT file starting at 0 */
-        long file_length;  /**< length of the file */
-        long file_position; /**< current position in the file */
+        uint64_t file_length;  /**< length of the file */
+        uint64_t file_position; /**< current position in the file */
 
         /* error counters */
         int32_t error_messages; /**< number of incomplete DLT messages found during file parsing */
@@ -590,22 +588,24 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
     } DltFile;
     """
 
-    _fields_ = [("handle", ctypes.POINTER(ctypes.c_int)),
-                ("index", ctypes.POINTER(ctypes.c_long)),
-                ("counter", ctypes.c_int32),
-                ("counter_total", ctypes.c_int32),
-                ("position", ctypes.c_int32),
-                ("file_length", ctypes.c_long),
-                ("file_position", ctypes.c_long),
-                ("error_messages", ctypes.c_int32),
-                ("filter", ctypes.POINTER(DLTFilter)),
-                ("filter_counter", ctypes.c_int32),
-                ("msg", DLTMessage)]
+    _fields_ = [
+        ("handle", ctypes.POINTER(ctypes.c_int)),
+        ("index", ctypes.POINTER(ctypes.c_long)),
+        ("counter", ctypes.c_int32),
+        ("counter_total", ctypes.c_int32),
+        ("position", ctypes.c_int32),
+        ("file_length", ctypes.c_uint64),
+        ("file_position", ctypes.c_uint64),
+        ("error_messages", ctypes.c_int32),
+        ("filter", ctypes.POINTER(DLTFilter)),
+        ("filter_counter", ctypes.c_int32),
+        ("msg", DLTMessage),
+    ]
 
     def __init__(self, **kwords):
         self.verbose = kwords.pop("verbose", 0)
         self.filename = kwords.pop("filename", None)
-        if six.PY3 and isinstance(self.filename, str):
+        if isinstance(self.filename, str):
             self.filename = bytes(self.filename, "utf-8")
         super(cDLTFile, self).__init__(**kwords)
         if dltlib.dlt_file_init(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
@@ -620,9 +620,9 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
 
     def __repr__(self):
         # pylint: disable=bad-continuation
-        return '<DLTFile object {} with {} messages>'.format(
-            "filename={}".format(self.filename) if self.filename else "",
-            self.counter_total)
+        return "<DLTFile object {} with {} messages>".format(
+            "filename={}".format(self.filename) if self.filename else "", self.counter_total
+        )
 
     def __del__(self):
         if dltlib.dlt_file_free(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
@@ -636,7 +636,7 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
         :rtype: int
         """
         with open(self.filename, "rb") as fobj:
-            last_position = self.file_position   # pylint: disable=access-member-before-definition
+            last_position = self.file_position  # pylint: disable=access-member-before-definition
             fobj.seek(last_position)
             buf = fobj.read(1024)
             while buf:
@@ -701,7 +701,7 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
         # load the filters
         self.set_filters(filters)
 
-        if six.PY3 and isinstance(filename, str):
+        if isinstance(filename, str):
             filename = bytes(filename, "utf-8")
         # read and index file
         self.filename = filename
@@ -713,11 +713,10 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
         if filters is not None:
             dlt_filter = DLTFilter(verbose=self.verbose)
             for apid, ctid in filters:
-                if six.PY3:
-                    if isinstance(apid, str):
-                        apid = bytes(apid, "ascii")
-                    if isinstance(ctid, str):
-                        ctid = bytes(ctid, "ascii")
+                if isinstance(apid, str):
+                    apid = bytes(apid, "ascii")
+                if isinstance(ctid, str):
+                    ctid = bytes(ctid, "ascii")
                 dlt_filter.add(apid, ctid)
             self.filters = dlt_filter
             dltlib.dlt_file_set_filter(ctypes.byref(self), ctypes.byref(dlt_filter), self.verbose)
@@ -768,7 +767,7 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
     def _open_file(self):
         """Open the configured file for processing"""
         file_opened = False
-        while not self.stop_reading.isSet():
+        while not self.stop_reading.is_set():
             if dltlib.dlt_file_open(ctypes.byref(self), self.filename, self.verbose) >= DLT_RETURN_OK:
                 file_opened = True
                 break
@@ -795,8 +794,9 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
     def __iter__(self):  # pylint: disable=too-many-branches
         """Iterate over messages in the file"""
         logger.debug("Starting File Read")
-        logger.debug("File Position: %d File Counter: %d File Name: %s",
-                     self.file_position, self.counter, self.filename)
+        logger.debug(
+            "File Position: %d File Counter: %d File Name: %s", self.file_position, self.counter, self.filename
+        )
         cached_mtime = 0
         cached_file_pos = 0
         corruption_check_try = True
@@ -804,7 +804,7 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
         self._open_file()
 
         found_data = False
-        while not self.stop_reading.isSet() or corruption_check_try:  # pylint: disable=too-many-nested-blocks
+        while not self.stop_reading.is_set() or corruption_check_try:  # pylint: disable=too-many-nested-blocks
             os_stat = os.stat(self.filename)
             mtime = os_stat.st_mtime
 
@@ -814,8 +814,10 @@ class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
 
                 while dltlib.dlt_file_read(ctypes.byref(self), self.verbose) >= DLT_RETURN_OK:
                     found_data = True
-                    if self.filter and dltlib.dlt_message_filter_check(
-                            ctypes.byref(self.msg), self.filter, 0) != DLT_RETURN_TRUE:
+                    if (
+                        self.filter
+                        and dltlib.dlt_message_filter_check(ctypes.byref(self.msg), self.filter, 0) != DLT_RETURN_TRUE
+                    ):
                         continue
 
                     index = self.position
@@ -863,6 +865,7 @@ class DLTClient(cDltClient):
     verbose = 0
 
     def __init__(self, **kwords):
+        self.is_udp_multicast = False
         self.verbose = kwords.pop("verbose", 0)
         if dltlib.dlt_client_init(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
             raise RuntimeError("Could not initialize DLTClient")
@@ -870,11 +873,29 @@ class DLTClient(cDltClient):
         if "servIP" in kwords:
             serv_ip = kwords.pop("servIP")
             if isinstance(serv_ip, str):
-                serv_ip = serv_ip.encode('utf8')
+                serv_ip = serv_ip.encode("utf8")
             ip_init_state = dltlib.dlt_client_set_server_ip(ctypes.byref(self), ctypes.create_string_buffer(serv_ip))
             if ip_init_state == DLT_RETURN_ERROR:
                 raise RuntimeError("Could not initialize servIP for DLTClient")
 
+            if ip.ip_address(serv_ip.decode("utf8")).is_multicast:
+                logger.info("Initializing DLTClient using UDP")
+                self.is_udp_multicast = True
+                if "hostIP" in kwords:
+                    host_ip = kwords.pop("hostIP")
+                    if isinstance(host_ip, str):
+                        host_ip = host_ip.encode("utf8")
+                    ip_init_state = dltlib.dlt_client_set_host_if_address(
+                        ctypes.byref(self), ctypes.create_string_buffer(host_ip)
+                    )
+                    if ip_init_state == DLT_RETURN_ERROR:
+                        raise RuntimeError("Could not initialize multicast address for DLTClient")
+
+                set_mode_state = dltlib.dlt_client_set_mode(ctypes.byref(self), DLT_CLIENT_MODE_UDP_MULTICAST)
+                logger.info("DLTClient using UDP set mode state: %s", set_mode_state)
+                if set_mode_state == DLT_RETURN_ERROR:
+                    raise RuntimeError("Could not initialize socket mode for DLTClient")
+
         # attribute to hold a reference to the connected socket in case
         # we created a connection with a timeout (via python, as opposed
         # to dltlib). This avoids the socket object from being garbage
@@ -887,11 +908,6 @@ class DLTClient(cDltClient):
         # it ourselves elsewhere
         self.port = kwords.get("port", DLT_DAEMON_TCP_PORT)
 
-    def __del__(self):
-        if dltlib.dlt_client_cleanup(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
-            raise RuntimeError("Could not cleanup DLTClient")
-        self.disconnect()
-
     def connect(self, timeout=None):
         """Connect to the server
 
@@ -905,39 +921,58 @@ class DLTClient(cDltClient):
         """
         connected = None
         error_count = 0
-        if timeout:
-            end_time = time.time() + timeout
-            while time.time() < end_time:
-                timeout_remaining = max(end_time - time.time(), 1) if timeout else None
-                try:
-                    self._connected_socket = socket.create_connection((ctypes.string_at(self.servIP), self.port),
-                                                                      timeout=timeout_remaining)
-                except IOError as exc:
-                    if error_count < MAX_LOG_IN_ROW:
-                        logger.debug("DLT client connect failed to connect to %s:%s : %s",
-                                     self.servIP, self.port, exc)
-                    error_count += 1
-                    time.sleep(1)
+        if not self.is_udp_multicast:
+            logger.info("Connecting DLTClient using TCP Connection")
+            if timeout:
+                end_time = time.time() + timeout
+                while time.time() < end_time:
+                    timeout_remaining = max(end_time - time.time(), 1) if timeout else None
+                    try:
+                        self._connected_socket = socket.create_connection(
+                            (ctypes.string_at(self.servIP), self.port), timeout=timeout_remaining
+                        )
+                    except IOError as exc:
+                        if error_count < MAX_LOG_IN_ROW:
+                            logger.debug(
+                                "DLT client connect failed to connect to %s:%s : %s", self.servIP, self.port, exc
+                            )
+                        error_count += 1
+                        time.sleep(1)
+
+                    if self._connected_socket:
+                        # pylint: disable=attribute-defined-outside-init
+                        self.sock = ctypes.c_int(self._connected_socket.fileno())
+                        # - also init the receiver to replicate
+                        # dlt_client_connect() behavior
+                        if hasattr(self.receiver, "type"):
+                            connected = dltlib.dlt_receiver_init(
+                                ctypes.byref(self.receiver), self.sock, DLT_RECEIVE_SOCKET, DLT_CLIENT_RCVBUFSIZE
+                            )
+                        else:
+                            connected = dltlib.dlt_receiver_init(
+                                ctypes.byref(self.receiver), self.sock, DLT_CLIENT_RCVBUFSIZE
+                            )
+                        break
+            else:
+                connected = dltlib.dlt_client_connect(ctypes.byref(self), self.verbose)
+                # - create a python socket object so that we can detect
+                # connection loss in the main_loop below as described at
+                # http://stefan.buettcher.org/cs/conn_closed.html
+                self._connected_socket = socket.fromfd(self.sock, socket.AF_INET6, socket.SOCK_STREAM)
+            if error_count > MAX_LOG_IN_ROW:
+                logger.debug("Surpressed %d messages for failed connection attempts", error_count - MAX_LOG_IN_ROW)
 
-                if self._connected_socket:
-                    # pylint: disable=attribute-defined-outside-init
-                    self.sock = ctypes.c_int(self._connected_socket.fileno())
-                    # - also init the receiver to replicate
-                    # dlt_client_connect() behavior
-                    connected = dltlib.dlt_receiver_init(ctypes.byref(self.receiver), self.sock, DLT_CLIENT_RCVBUFSIZE)
-                    break
         else:
+            logger.info("Connecting DLTClient using UDP Connection")
             connected = dltlib.dlt_client_connect(ctypes.byref(self), self.verbose)
-            # - create a python socket object so that we can detect
-            # connection loss in the main_loop below as described at
-            # http://stefan.buettcher.org/cs/conn_closed.html
-            self._connected_socket = socket.fromfd(self.sock, socket.AF_INET6, socket.SOCK_STREAM)
-        if error_count > MAX_LOG_IN_ROW:
-            logger.debug("Surpressed %d messages for failed connection attempts", error_count - MAX_LOG_IN_ROW)
+
+        logger.info("DLT Connection return: %s", connected)
         return connected == DLT_RETURN_OK
 
     def disconnect(self):
         """Close all sockets"""
+        if dltlib.dlt_client_cleanup(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
+            raise RuntimeError("Could not cleanup DLTClient")
         if self._connected_socket:
             try:
                 self._connected_socket.shutdown(socket.SHUT_RDWR)
@@ -960,11 +995,13 @@ class DLTClient(cDltClient):
         :rtype: DLTMessage|None
         """
         msg = DLTMessage(verbose=verbose)
-        res = dltlib.dlt_message_read(ctypes.byref(msg),
-                                      ctypes.cast(self.receiver.buf, ctypes.POINTER(ctypes.c_uint8)),
-                                      ctypes.c_uint(self.receiver.bytesRcvd),  # length
-                                      ctypes.c_int(0),  # resync
-                                      ctypes.c_int(verbose))  # verbose
+        res = dltlib.dlt_message_read(
+            ctypes.byref(msg),
+            ctypes.cast(self.receiver.buf, ctypes.POINTER(ctypes.c_uint8)),
+            ctypes.c_uint(self.receiver.bytesRcvd),  # length
+            ctypes.c_int(0),  # resync
+            ctypes.c_int(verbose),
+        )  # verbose
 
         if res != DLT_MESSAGE_ERROR_OK:
             # - failed to read a complete message, possibly read an incomplete
@@ -985,6 +1022,25 @@ class DLTClient(cDltClient):
         """Get the mode"""
         return getattr(self, "mode", getattr(super(DLTClient, self), "serial_mode", 0))
 
+    @ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(DLTMessage), ctypes.c_void_p)
+    def msg_callback(msg, data):  # pylint: disable=no-self-argument
+        """Implements a simple callback that prints a dlt message received"""
+        if msg is None:
+            print("NULL message in callback")
+            return -1
+        if msg.contents.p_standardheader.contents.htyp & DLT_HTYP_WEID:
+            dltlib.dlt_set_storageheader(msg.contents.p_storageheader, msg.contents.headerextra.ecu)
+        else:
+            dltlib.dlt_set_storageheader(msg.contents.p_storageheader, ctypes.c_char_p(""))
+
+        print(msg.contents)
+        return 0
+
+    def client_loop(self):
+        """Executes native dlt_client_main_loop() after registering msg_callback method as callback"""
+        dltlib.dlt_client_register_message_callback(self.msg_callback)
+        dltlib.dlt_client_main_loop(ctypes.byref(self), None, self.verbose)
+
 
 # pylint: disable=too-many-arguments,too-many-return-statements,too-many-branches
 def py_dlt_client_main_loop(client, limit=None, verbose=0, dumpfile=None, callback=None):
@@ -1001,17 +1057,18 @@ def py_dlt_client_main_loop(client, limit=None, verbose=0, dumpfile=None, callba
         # this will raise a socket.timeout exception which the caller is
         # expected to handle (possibly by attempting a reconnect)
         # pylint: disable=protected-access
-        try:
-            ready_to_read = client._connected_socket.recv(1, socket.MSG_PEEK | socket.MSG_DONTWAIT)
-        except OSError as os_exc:
-            logger.error("[%s]: DLTLib closed connected socket", os_exc)
-            return False
+        if not client.is_udp_multicast:
+            try:
+                ready_to_read = client._connected_socket.recv(1, socket.MSG_PEEK | socket.MSG_DONTWAIT)
+            except OSError as os_exc:
+                logger.error("[%s]: DLTLib closed connected socket", os_exc)
+                return False
 
-        if not ready_to_read:
-            # - implies that the other end has called close()/shutdown()
-            # (ie: clean disconnect)
-            logger.debug("connection terminated, returning")
-            return False
+            if not ready_to_read:
+                # - implies that the other end has called close()/shutdown()
+                # (ie: clean disconnect)
+                logger.debug("connection terminated, returning")
+                return False
 
         # - check if stop flag has been set (end of loop)
         if callback and not callback(None):
@@ -1024,7 +1081,10 @@ def py_dlt_client_main_loop(client, limit=None, verbose=0, dumpfile=None, callba
         # the status of the callback (in the case of dlt_broker, this is
         # the stop_flag Event), this loop will only proceed after the
         # function has returned or terminate when an exception is raised
-        recv_size = dltlib.dlt_receiver_receive(ctypes.byref(client.receiver), DLT_RECEIVE_SOCKET)
+        if hasattr(client.receiver, "type"):
+            recv_size = dltlib.dlt_receiver_receive(ctypes.byref(client.receiver))
+        else:
+            recv_size = dltlib.dlt_receiver_receive(ctypes.byref(client.receiver), DLT_RECEIVE_SOCKET)
         if recv_size <= 0:
             logger.error("Error while reading from socket")
             return False
